#include "system.h"

static Texture* clOutput = 0;
static Shader* shader = 0;
static Kernel* clearKernel = 0, *renderKernel = 0, *updateKernel = 0, *resetKernel = 0, *gridKernel = 0;
static Kernel* sim1Kernel = 0, *sim2Kernel = 0, *simXKernel = 0;
static Buffer* outputBuffer, *balls, *balls2, *grid;
static float gr = 0;

bool Game::Init()
{
	// load shader and texture
	clOutput = new Texture( SCRWIDTH, SCRHEIGHT, Texture::FLOAT );
	shader = new Shader( "shaders/vignette.vert", "shaders/vignette.frag" );
	// load OpenCL code
	clearKernel = new Kernel( "programs/program.cl", "clear" );
	renderKernel = new Kernel( "programs/program.cl", "render" );
	updateKernel = new Kernel( "programs/program.cl", "update" );
	resetKernel = new Kernel( "programs/program.cl", "clearGrid" );
	gridKernel = new Kernel( "programs/program.cl", "fillGrid" );
	sim1Kernel = new Kernel( "programs/program.cl", "simulate1" );
	sim2Kernel = new Kernel( "programs/program.cl", "simulate2" );
	simXKernel = new Kernel( "programs/program.cl", "simulateX" );
	// link cl output texture as an OpenCL buffer
	outputBuffer = new Buffer( clOutput->GetID(), Buffer::TARGET );
	balls = new Buffer( BALLCOUNT * 6 * sizeof( float ) );
	balls2 = new Buffer( BALLCOUNT * 6 * sizeof( float ) );
	grid = new Buffer( GRIDX * GRIDY * (BALLSPERCELL + 1) * sizeof( unsigned int ) );
	clearKernel->SetArgument( 0, outputBuffer );
	renderKernel->SetArgument( 0, outputBuffer );
	renderKernel->SetArgument( 1, balls );
	renderKernel->SetArgument( 2, grid );
	resetKernel->SetArgument( 0, grid );
	gridKernel->SetArgument( 0, balls );
	gridKernel->SetArgument( 1, grid );
	sim1Kernel->SetArgument( 0, balls );
	sim1Kernel->SetArgument( 1, grid );
	sim2Kernel->SetArgument( 0, balls );
	sim2Kernel->SetArgument( 1, balls2 );
	sim2Kernel->SetArgument( 2, grid );
	simXKernel->SetArgument( 0, balls );
	simXKernel->SetArgument( 1, balls2 );
	simXKernel->SetArgument( 2, grid );
	// put initial ball positions in buffer
	float* fb = (float*)balls->GetHostPtr();
	for( int i = 0; i < BALLCOUNT; i++ )
	{
		fb[i * 6] = Rand( 1 );
		fb[i * 6 + 1] = Rand( 1 );
		fb[i * 6 + 2] = Rand( 0.01f ) - 0.005f;
		fb[i * 6 + 3] = Rand( 0.01f ) - 0.005f;
		fb[i * 6 + 4] = fb[i * 6 + 0];
		fb[i * 6 + 5] = fb[i * 6 + 1];
	}
	balls->CopyToDevice();
	// done
	return true;
}

void Game::Tick()
{
	// initialize
	clearKernel->Run( outputBuffer );
	// simulate stage 1
	gr += 0.005f;
	if (gr >= 2 * PI) gr -= 2 * PI;
	float fx = 0.0002f * sinf( gr );
	float fy = 0.0002f * cosf( gr );
	sim1Kernel->SetArgument( 1, fx );
	sim1Kernel->SetArgument( 2, fy );
	sim1Kernel->Run( BALLCOUNT );
	// update grid
	resetKernel->Run( GRIDX * GRIDY );
	gridKernel->Run( BALLCOUNT );
	// simulate stage 2
	balls->CopyTo( balls2 );
#if 1
	sim2Kernel->Run( GRIDX * GRIDY );
#else
	simXKernel->Run( BALLCOUNT );
#endif
	// update grid
	resetKernel->Run( GRIDX * GRIDY );
	gridKernel->Run( BALLCOUNT );
	// run cl code to fill texture
	renderKernel->Run( outputBuffer );
	// run shader on cl-generated texture
	shader->Bind();
	shader->SetInputTexture( GL_TEXTURE0, "color", clOutput );
	shader->SetInputMatrix( "view", mat4::Identity() );
	DrawQuad();
}

void Game::Shutdown()
{
	delete clOutput;
	delete shader;
	delete renderKernel;
	delete updateKernel;
}

// EOF